<?php
/**
 * The zen file of thinkrun module of ZenTaoPMS.
 *
 * @copyright   Copyright 2009-2024 禅道软件（青岛）有限公司(ZenTao Software (Qingdao) Co., Ltd. www.zentao.net)
 * @license     ZPL(https://zpl.pub/page/zplv12.html) or AGPL(https://www.gnu.org/licenses/agpl-3.0.en.html)
 * @author      Gang Liu <liugang@easycorp.ltd>
 * @package     thinkrun
 * @link        https://www.zentao.net
 */
class thinkrunZen extends thinkrun
{
    /**
     * 设置分析的导航菜单。
     * Set menu of thinkrun module.
     *
     * @param  int $marketID
     * @access public
     * @return void
     */
    public function setMenu($marketID = 0)
    {
        if($marketID) $this->loadModel('market')->setMenu($marketID);
        if($this->app->rawMethod == 'answer' || $this->app->rawMethod == 'result')
        {
            if(isset($this->lang->market->homeMenu->wizard['subMenu'])) unset($this->lang->market->homeMenu->wizard['subMenu']);
            if(isset($this->lang->market->menu->wizard['subMenu']))     unset($this->lang->market->menu->wizard['subMenu']);
        }
    }

    /**
     * 获取上一步的跳转连接：如果有上一步骤则跳转到上一步骤，否则返回空字符串。
     * Get the redirect URL of the previous step: if there is a previous step, redirect to the previous step, otherwise return an empty string.
     *
     * @param  array  $steps
     * @param  int    $marketID
     * @param  int    $stepID
     * @param  string $browseType
     * @access public
     * @return string
     */
    public function getPrevURL($steps, $marketID, $stepID, $browseType)
    {
        $prevStep = $this->loadModel('thinkstep')->getStep($steps, $stepID, 'prev');

        if($prevStep)
        {
            $prevStep->options = is_string($prevStep->options) ? json_decode($prevStep->options) : $prevStep->options;

            /* 上一题为多选题或者多列填空题并且引用了其他选项。The previous question is a multiple-choice or multicolumn question and cites other options. */
            if($prevStep->type == 'question' && in_array($prevStep->options->questionType, $this->config->thinkstep->quoteQuestionType) && !empty($prevStep->options->setOption))
            {
                /* 上一题没有选项的情况，更新上一题的hidden字段隐藏该题。*/
                /* If there are no options for the previous question, update the hidden field of the next question to hide it. */
                $emptyCheckboxFields = empty($this->getCheckboxFields($steps, $prevStep));
                foreach($steps as $key => $item)
                {
                    if($item->id == $prevStep->id)
                    {
                        $oldStep             = clone $item;
                        $steps[$key]->hidden = (int)$emptyCheckboxFields;
                        $item->hidden        = (int)$emptyCheckboxFields;

                        /* 删除掉数据库中不存在的字段。Delete fields that do not exist in the database.*/
                        unset($oldStep->url, $oldStep->index, $item->url, $item->index);
                        $this->thinkstep->update($item, $oldStep, true);
                        break;
                    }
                }
                if($emptyCheckboxFields)  $prevStepURL = $this->getPrevURL($steps, $marketID, $prevStep->id, $browseType);
                if(!$emptyCheckboxFields) $prevStepURL = inlink('answer', "marketID=$marketID&stepID=$prevStep->id&browseType=$browseType");
            }
            else
            {
                $prevStepURL = inlink('answer', "marketID=$marketID&stepID=$prevStep->id&browseType=$browseType");
            }
            return $prevStepURL;
        }
        return '';
    }

    /**
     * 获取下一步的跳转连接：如果有下一步骤则跳转到下一步骤，否则跳转到结果展示。
     * Get the redirect URL: if there is a next step, redirect to the next step, otherwise redirect to the result page.
     *
     * @param  array  $steps
     * @param  int    $marketID
     * @param  int    $stepID
     * @param  string $browseType  浏览类型。all 全部 | doing 进行中 | done 已完成
     * @access public
     * @return string
     */
    public function getNextURL($steps, $marketID, $stepID, $browseType = 'doing')
    {
        $nextStep    = $this->loadModel('thinkstep')->getStep($steps, $stepID);
        $nextStepURL = '';

        if($nextStep && empty($nextStepURL))
        {
            $nextStep->options = json_decode($nextStep->options);
            /* 下一题为多选题或多列填空题并且引用了其他选项。The next question is a multiple-choice or multicolumn question and cites other options. */
            if($nextStep->type == 'question' && in_array($nextStep->options->questionType, $this->config->thinkstep->quoteQuestionType) && !empty($nextStep->options->setOption))
            {
                /* 下一题没有选项的情况，更新下一题的hidden字段隐藏该题。*/
                /* If there are no options for the next question, update the hidden field of the next question to hide it. */
                $emptyCheckboxFields = empty($this->getCheckboxFields($steps, $nextStep));
                foreach($steps as $key => $item)
                {
                    if($item->id == $nextStep->id)
                    {
                        $oldStep             = clone $item;
                        $steps[$key]->hidden = (int)$emptyCheckboxFields;
                        $item->hidden        = (int)$emptyCheckboxFields;

                        /* 删除掉数据库中不存在的字段。Delete fields that do not exist in the database.*/
                        unset($oldStep->url, $oldStep->index, $item->url, $item->index);
                        $this->thinkstep->update($item, $oldStep, true);
                        break;
                    }
                }
                if($emptyCheckboxFields)  $nextStepURL = $this->getNextURL($steps, $marketID, $nextStep->id, $browseType);
                if(!$emptyCheckboxFields) $nextStepURL = inlink('answer', "marketID=$marketID&stepID=$nextStep->id&browseType=$browseType");
            }
            else
            {
                $nextStepURL = inlink('answer', "marketID=$marketID&stepID=$nextStep->id&browseType=$browseType");
            }
        }
        else
        {
            $step        = $this->thinkstep->getByID($stepID);
            $nextStepURL = inlink('result', "marketID=$marketID&runID=$step->run&browseType=$browseType");
        }
        return $nextStepURL;
    }

    /**
     * 更新一个分析的当前步骤。
     * Update the current step of a thinkrun.
     *
     * @param  array   $steps
     * @param  object  $step
     * @access public
     * @return bool
     */
    public function updateCurrentStep($steps, $step)
    {
        $nextStep = $this->thinkstep->getStep($steps, $step->id);
        if($nextStep) $this->thinkrun->updateCurrentStep($step->run, $nextStep->id);
        return !dao::isError();
    }

    /**
     * 检查表格填空题的表单数据是否符合要求
     * Check if the form data of the table input question meets the requirements.
     *
     * @param  object  $options 后台设置的配置项
     * @param  array   $answer  用户自己输入的内容
     * @param  object  $wizard
     * @param  int     runID
     * @access public
     * @return array|false
     */
    public function checkTableInput($step, $answer, $wizard, $runID)
    {
        $options = $step->options;

        /* 输入金额的数字类型需要将字符串两端的0去除掉,为了保证小数点的精度和数字的有效性。*/
        /* The numerical type of the input amount needs to remove the zeros at both ends of the string to ensure the accuracy of the decimal point and the validity of the numbers. */
        if(!empty($options->inputType) && $options->inputType == '2')
        {
            foreach ($answer->result as &$result){$result = trim(preg_replace('/^0+(?=\d)|0+\s*$/', '', $result));}
        }
        /* 检查用户回答的后台设置的问题是否满足必填行数的要求。Check if the user's answers meet the requirements of the required number of rows. */
        if(!empty($options->required))
        {
            $hasResultCount = 0;
            foreach($options->fields as $key => $field)
            {
                $resultItem = !empty($answer->result[$key]) ? $answer->result[$key] : '';
                if(!empty($options->inputType) && $options->inputType == '2' && !empty($resultItem)) $resultItem = (float) $resultItem;
                if(!empty($resultItem)) $hasResultCount++;
            }
            if($hasResultCount < $options->requiredRows) return array('result' => 'fail', 'callback' => "
                $('.run-error-msg').removeClass('hidden');
                const errorElement = $('.run-error-msg');
                if(errorElement && errorElement[0]) errorElement[0].scrollIntoView({ behavior: 'auto', block: 'center' });
            ");
        }

        $answer->customFields = !empty($answer->customFields) ? $answer->customFields : array();
        /* 检查用户自己添加的问题标题是否重复。Check if the custom fields are repeated. */
        $customFieldsCount = array_count_values($answer->customFields);
        foreach($answer->customFields as $key => $field)
        {
            if($field && (in_array($field, $options->fields) || $customFieldsCount[$field] > 1)) return array('result' => 'fail', 'message' => array("customFields[$key]" => $this->lang->thinkrun->error->repeatRowTitle));
        }

        /* 检查用户自己添加的问题标题是否为空。Check if the custom fields are empty. */
        foreach($answer->result as $key => $value)
        {
            if(isset($options->fields[$key])) continue;
            if($value && empty($answer->customFields[$key])) return array('result' => 'fail', 'message' => array("customFields[$key]" => $this->lang->thinkrun->error->emptyRowTitle));
        }

        if(!empty($options->inputType) && $options->inputType == '1')
        {
            $sum = 0;
            foreach ($answer->result as $result)
            {
                $sum += $result;
            }
            if($sum != 100) return $this->send(array('result' => 'fail', 'callback' => "zui.Modal.alert({message: '{$this->lang->thinkrun->error->weight}', icon: 'icon-exclamation-sign', iconClass: 'warning-pale rounded-full icon-2x'})"));
        }

        if($wizard->model == 'bcg' && empty($options->inputType))
        {
            /* 波士顿矩阵模型第2、3题的输入项不能重复。The input items in Boston Matrix Model Questions 2 and 3 cannot be repeated. */
            $resultCounts = array_count_values($answer->result);
            foreach($answer->result as $resultIndex => $result)
            {
                if($resultCounts[$result] > 1) return array('result' => 'fail', 'message' => array("result[$resultIndex]" => $this->lang->thinkrun->error->repeatContent), 'callback' => "$('.run-error-msg').addClass('hidden');");
            }
            /* 第3题的输入项不能和第1题的输入项重复。The input items for question 3 cannot be the same as the input items for question 1. */
            $children = $this->getFirstNodeQuestion($wizard->id, 0);
            if($step->origin == $children[2]->id)
            {
                $intersect     = array_intersect($options->fields, $answer->result);
                $intersectKeys = array_keys(array_intersect($answer->result, $intersect));
                if(!empty($intersectKeys)) return array('result' => 'fail', 'message' => array('result[' . $intersectKeys[0] . ']' => $this->lang->thinkrun->error->repeatProduct), 'callback' => "$('.run-error-msg').addClass('hidden');");
            }
        }
        return false;
    }

    /**
     * 处理时间字段。
     * Process time fields.
     *
     * @param  array  $runs
     * @access public
     * @return array
     */
    public function processTimeFields($runs)
    {
        foreach($runs as $run)
        {
            if($run->estimate) $run->estimate .= $this->config->thinkrun->minute;
            if($run->consumed) $run->consumed  = round($run->consumed / 60, 2) . $this->config->thinkrun->minute;
        }
        return $runs;
    }

    /**
     * 构建结果页图形相关数据。
     * Build result page graphics data.
     *
     * @param  int    $runID
     * @param  object $wizard
     * @param  array  $steps
     * @param  array  $stepTitles
     * @access public
     * @return void
     */
    public function buildResultByGraph($runID, $wizard, $steps, $stepTitles)
    {
        $previewBlocks = array();
        foreach($wizard->blocks as $block)
        {
            $origin    = (int)$this->loadModel('thinkstep')->getIDByOrigin((int)$block, $runID);
            $questions = array_filter($steps, function($stepItem) use ($block, $origin)
            {
                $path = explode(',', trim($stepItem->path, ','));
                if(is_string($stepItem->options)) $stepItem->options = json_decode($stepItem->options);
                if(is_string($stepItem->answer))  $stepItem->answer  = json_decode($stepItem->answer);
                return in_array($block, $path) || in_array($origin, $path);
            });

            $item = new stdClass();
            $item->text  = !isset($stepTitles[$block]) ? '' : $stepTitles[$block];
            $item->id    = $block;
            $item->steps = $questions;
            $previewBlocks[] = $item;
        }
        $this->view->previewBlocks = $previewBlocks;
    }

    /**
     * 构建韦恩模型结果页相关数据。
     * Build result data for Venn model.
     *
     * @param  object $wizard
     * @param  array  $questionSteps
     * @access public
     * @return void
     */
    public function buildResultDataForVenn($wizard, $questionSteps)
    {
        $previewBlocks = array();
        foreach($wizard->blocks as $index => $block)
        {
            $blockSteps = $this->loadModel('thinkwizard')->getRelatedStepsByBlock($wizard->id, $index + 1, $questionSteps);

            $item = new stdClass();
            $item->text  = $block;
            $item->id    = $index;
            $item->steps = $blockSteps;
            $previewBlocks[] = $item;
        }
        $this->view->previewBlocks = $previewBlocks;
    }

    /**
     * 处理多列填空题导出表格的相关数据。
     * Process multiple column fill in the blank questions and export relevant data from the table.
     *
     * @param  object $step           多列填空题分析步骤
     * @param  array  $rows           导出表格的数据
     * @param  array  $rowspan        导出表格合并列的数据
     * @param  array  $colspan        导出表格合并行的数据
     * @param  int    $stepIndex      分析步骤结果的序号
     * @param  string $block          分析步骤关联的区块
     * @param  int    $stepsNum       分析步骤总的单元格行数
     * @access public
     * @return array
     */
    public function processMulticolumn($step, $rows, $rowspan, $colspan, $stepIndex = 0, $block = '', $stepsNum = 0)
    {
        $stepOptions     = is_string($step->options) ? json_decode($step->options) : $step->options;
        $stepAnswer      = is_string($step->answer) ? json_decode($step->answer) : $step->answer;
        $result          = (array) $stepAnswer->result;
        $multicolumnData = array();
        $stepTemp        = new stdClass;
        $resultMaxCount  = count((array) reset($result));
        $resultCount     = count((array) reset($result));

        foreach($stepOptions->fields as $fieldIndex => $filed)
        {
            $answerResultName            = "answerResult" . ($fieldIndex + 1);
            $stepTemp->index             = $stepIndex;
            $stepTemp->block             = $block;
            $stepTemp->node              = !empty($step->node) ? $step->node : '';
            $stepTemp->title             = $step->title;
            $stepTemp->questionType      = $step->questionType;
            $stepTemp->$answerResultName = $filed;
        }
        $rows[] = $stepTemp;
        for($i = 1; $i <= $resultMaxCount; $i++)
        {
            $stepTemp    = new stdClass;
            $emptyResult = true;

            foreach($stepOptions->fields as $fieldIndex => $filed)
            {
                $index                       = $fieldIndex + 1;
                $columnName                  = 'col' . $index;
                $answerResultName            = "answerResult" . ($fieldIndex + 1);
                $answerResultValue           = $result[$columnName]->$i;
                $emptyResult                 = !empty($answerResultValue) ? false : $emptyResult;
                $stepTemp->index             = $stepIndex;
                $stepTemp->title             = $step->title;
                $stepTemp->node              = !empty($step->node) ? $step->node : '';
                $stepTemp->questionType      = $step->questionType;
                $stepTemp->$answerResultName = !empty($answerResultValue) ? $answerResultValue : $this->lang->thinkrun->tableResult->notFilledIn;
            }

            if(!empty($emptyResult)) $resultCount = $resultCount - 1;
            if(empty($emptyResult))  $rows[] = $stepTemp;
        }
        if($resultCount >= 1)
        {
            $rowspanNum = $resultCount + 1;
            $rowspan[count($rows) - $resultCount - 1] = array
            (
                'rows' => 'block,index,node,questionType,title',
                'num'  => $rowspanNum
            );
        }
        $multicolumnData['rows']    = $rows;
        $multicolumnData['rowspan'] = $rowspan;
        $multicolumnData['colspan'] = $colspan;
        return $multicolumnData;
    }

    /**
     * 处理表格填空表格的相关数据。
     * Process the relevant data of fill in the blanks in the table.
     *
     * @param  object $step            表格填空题分析步骤
     * @param  int    $maxColumnResult 表格结果最大数
     * @param  array  $rows            导出表格的数据
     * @param  array  $rowspan         导出表格合并列的数据
     * @param  array  $colspan         导出表格合并行的数据
     * @param  int    $stepIndex       分析步骤结果的序号
     * @param  string $block           分析步骤关联的区块
     * @access public
     * @return array
     */
    public function processTableInput($step, $maxColumnResult, $rows, $rowspan, $colspan, $stepIndex = 0, $block = '')
    {

        $tableInputData = array();
        $stepOptions    = is_string($step->options) ? json_decode($step->options) : $step->options;
        $stepAnswer     = is_string($step->answer) ? json_decode($step->answer) : $step->answer;
        $stepResult     = (array) $stepAnswer->result;
        $stepAmountUnit = !empty($stepAnswer->amountUnit) ? $stepAnswer->amountUnit : array();
        $customFields   = !empty($stepAnswer->customFields) ? get_object_vars($stepAnswer->customFields) : array();
        $newFields      = array_merge($stepOptions->fields, $customFields);
        $fieldsCount    = count($newFields);

        if($fieldsCount > 1)
        {
            $rowspan[count($rows)] = array
            (
                'rows' => 'block,index,node,questionType,title',
                'num'  => $fieldsCount
            );
        }
        foreach($newFields as $newIndex => $item)
        {
            if($maxColumnResult > 1) $colspan[count($rows)] = array('cols' => 'answerResult2', 'num' => $maxColumnResult - 1);
            $answerResult2 = isset($stepResult[$newIndex]) ? $stepResult[$newIndex] : null;
            $title         = $step->title;
            if(!empty($stepOptions->inputType) && !empty($answerResult2)) $answerResult2 = $answerResult2 . ($stepOptions->inputType == '1' ? '%' : $this->lang->thinkrun->amountUnit[$stepAmountUnit[$newIndex]]);
            if($stepOptions->questionType == 'score')
            {
                if(!empty($answerResult2)) $answerResult2 = $answerResult2 . $this->lang->thinkrun->tableResult->points;
                if(isset($stepOptions->scoreSetting)) $title = $title . ' （' . $this->lang->thinkwizard->questions->scoreList[$stepOptions->scoreSetting] . '）';
            }

            $stepTemp                = new stdClass;
            $stepTemp->block         = $block;
            $stepTemp->index         = $stepIndex;
            $stepTemp->node          = !empty($step->node) ? $step->node : '';
            $stepTemp->title         = $title;
            $stepTemp->questionType  = $step->questionType;
            $stepTemp->answerResult1 = $item;
            $stepTemp->answerResult2 = !empty($answerResult2) ? $answerResult2 : $this->lang->thinkrun->tableResult->notFilledIn;
            $rows[]                  = $stepTemp;
        }
        $tableInputData['rows']    = $rows;
        $tableInputData['rowspan'] = $rowspan;
        $tableInputData['colspan'] = $colspan;
        return $tableInputData;
    }

    /**
     * 处理韦恩图模型的导出表格相关数据。
     * Processing export table related data for Wayne diagram model.
     *
     * @param  int    $blockIndex      关联区块的ID
     * @param  string $block           关联区块的自定义名称
     * @param  object $step            多列填空题分析步骤
     * @param  int    $maxColumnResult 结果表格的最大列数
     * @param  array  $rows            导出表格的数据
     * @param  array  $rowspan         导出表格合并列的数据
     * @param  array  $colspan         导出表格合并行的数据
     * @access public
     * @return array
     */
    public function processVennTableRows($blockIndex, $block, $steps, $maxColumnResult, $rows, $rowspan, $colspan, $showDefaltBlock=false)
    {
        $tableData            = array();
        $letter               = range('a', 'z');
        $blockName            = $block . ($showDefaltBlock ? ' (' . $letter[$blockIndex -1] . $this->lang->thinkrun->tableResult->block . ')' : '');
        $tableData['rows']    = $rows;
        $tableData['rowspan'] = $rowspan;
        $tableData['colspan'] = $colspan;

        if(empty($steps))
        {
            $step = new stdClass;
            $step->block = $blockName;
            $step->questionType = $this->lang->thinkrun->tableResult->noneContent;

            $tableData['rows'][]                = $step;
            $tableData['colspan'][count($rows)] = array('cols' => 'questionType', 'num' => $maxColumnResult + 2);
            return $tableData;
        }
        foreach($steps as $stepItem)
        {
            $step        = clone $stepItem;
            $stepLink    = json_decode($step->link);
            $stepOptions = $step->options;
            $stepAnswer  = $step->answer;
            $result      = (array) $stepAnswer->result;
            $stepsCount  = count($steps);

            $step->questionType = $this->lang->thinkrun->tableResult->tableQuestionType[$stepOptions->questionType];
            if($stepOptions->questionType === 'tableInput')
            {
                $tableInputData = $this->processTableInput($step, $maxColumnResult, $rows, $rowspan, $colspan, 0, $blockName);
                $rows           = $tableInputData['rows'];
                $rowspan        = $tableInputData['rowspan'];
                $colspan        = $tableInputData['colspan'];
            }
            elseif($stepOptions->questionType == 'multicolumn')
            {
                $multicolumnData = $this->processMulticolumn($step, $rows, $rowspan, $colspan, 0, $blockName, $stepsCount);
                $rows            = $multicolumnData['rows'];
                $rowspan         = $multicolumnData['rowspan'];
                $colspan         = $multicolumnData['colspan'];
            }
            else
            {
                if($maxColumnResult > 1) $colspan[count($rows)] = array('cols' => 'answerResult1', 'num' => $maxColumnResult);
                $resultValue = '';
                if(!empty($stepLink->unSelectedBlock) && $stepLink->unSelectedBlock == $blockIndex)
                {
                    $unSelectedItems = array_diff($stepOptions->fields, $result);
                    foreach($unSelectedItems as $unSelectedItem)
                    {
                        $resultValue = $resultValue . $unSelectedItem . ' ';
                    }
                    if(isset($stepOptions->enableOther) && $stepOptions->enableOther === 'on' && empty($stepAnswer->other)) $resultValue = $resultValue . $this->lang->thinkrun->other;
                }
                else
                {
                    foreach($result as $item)
                    {
                        $resultValue = $resultValue . ($item === 'other' ? $stepAnswer->other : $item) . '  ';
                    }
                }
                $step->index         = 0;
                $step->block         = $blockName;
                $step->answerResult1 = !empty($resultValue) ? $resultValue : $this->lang->thinkrun->tableResult->notFilledIn;
                $rows[]              = $step;
            }
        }
        $tableData['rows']    = $rows;
        $tableData['rowspan'] = $rowspan;
        $tableData['colspan'] = $colspan;
        return $tableData;
    }


    /**
     * 处理导出表格数据。
     * Process export table datas.
     *
     * @param  array  $questionSteps
     * @access public
     * @return array
     */
    public function processTableRows($questionSteps, $maxColumnResult)
    {
        $stepIndex = 1;
        $tableData = array();
        $rows      = array();
        $rowspan   = array();
        $colspan   = array();

        foreach($questionSteps as $step)
        {
            $stepOptions = json_decode($step->options);
            $stepAnswer  = json_decode($step->answer);
            $stepResult  = (array) $stepAnswer->result;

            $step->questionType = $this->lang->thinkrun->tableResult->tableQuestionType[$stepOptions->questionType];

            if($stepOptions->questionType === 'tableInput' || $stepOptions->questionType === 'score')
            {
                $tableInputData = $this->processTableInput($step, $maxColumnResult, $rows, $rowspan, $colspan, $stepIndex);
                $rows           = $tableInputData['rows'];
                $rowspan        = $tableInputData['rowspan'];
                $colspan        = $tableInputData['colspan'];
            }
            elseif($stepOptions->questionType == 'multicolumn')
            {
                $multicolumnData = $this->processMulticolumn($step, $rows, $rowspan, $colspan, $stepIndex);
                $rows            = $multicolumnData['rows'];
                $rowspan         = $multicolumnData['rowspan'];
                $colspan         = $multicolumnData['colspan'];
            }
            else
            {
                if($maxColumnResult > 1) $colspan[count($rows)] = array('cols' => 'answerResult1', 'num' => $maxColumnResult);
                $result = '';
                foreach($stepResult as $item)
                {
                    $result = $result . ($item === 'other' ? $stepAnswer->other : $item) . '  ';
                }
                $step->index         = $stepIndex;
                $step->answerResult1 = !empty($result) ? $result : $this->lang->thinkrun->tableResult->notFilledIn;
                $rows[]              = $step;
            }
            $stepIndex ++;
        }

        $tableData['rows']      = $rows;
        $tableData['rowspan']   = $rowspan;
        $tableData['colspan']   = $colspan;
        return $tableData;
    }

    /**
     * 获取引用其他问题选项的多选题的选项列表。
     * Get fields for multiple choice questions that reference other question options.
     *
     * @param  array  $steps
     * @param  object $step
     * @access public
     * @return array
     */
    public function getCheckboxFields ($steps, $step)
    {
        $fields     = array();
        $quoteTitle = explode(', ', $step->options->quoteTitle);

        foreach($steps as $item)
        {
            if(in_array($item->origin, $quoteTitle))
            {
                $index         = array_search($item->origin, $quoteTitle);
                $citations     = explode(', ', $step->options->citation);
                $selectColumns = !empty($step->options->selectColumn) ? explode(', ', $step->options->selectColumn) : array();
                $itemAnswer    = !empty($item->answer) ? json_decode($item->answer) : new stdClass();
                $itemOptions   = is_string($item->options) ? json_decode($item->options) : $item->options;
                $itemOptions->fields = !empty($itemOptions->fields) ? $itemOptions->fields : array();
                $itemAnswer->result  = !empty($itemAnswer->result) ? $itemAnswer->result : array();
                $quoteFields = $itemOptions->fields;
                if($citations[$index] == 0) $fields = $quoteFields;
                if($citations[$index] == 1) $fields = array_merge($fields, array_intersect($itemOptions->fields, $itemAnswer->result));
                if($citations[$index] == 2) $fields = array_merge($fields, array_diff($itemOptions->fields, $itemAnswer->result));
                if($citations[$index] != 2 && $citations[$index] != 3 && !empty($itemAnswer->result) && in_array('other',$itemAnswer->result)) $fields[] = $itemAnswer->other ? $itemAnswer->other : $this->lang->other;
                if($citations[$index] == 3)
                {
                    /* 过滤掉多列填空的空值，并且重新从0排序。Filter out empty values in multiple columns and reorder from 0。 */
                    $colNum      = 'col' . $selectColumns[$index];
                    $fieldResult = array_values(array_filter((array) $itemAnswer->result->$colNum));
                    $fields      = array_merge($fields, $fieldResult);
                }
                if($citations[$index] == 4) $fields = array_merge($fields, $itemAnswer->result);
                if($citations[$index] == 5) $fields = array_merge($fields, $itemAnswer->result);
            }
        }
        return $fields;
    }

    /**
     * 校验多列填空题的答案。
     * Check the answers to multiple column.
     *
     * @param  object $options
     * @param  object $answer
     * @param  object $wizard
     * @param  int    $runID
     * @access public
     * @return array|bool
     */
    public function checkMulticolumn($options, $answer, $wizard, $runID)
    {
        if(!empty($options->required))
        {
            foreach($options->requiredCols as $requiredCol)
            {
                foreach($answer->result["col$requiredCol"] as &$result)
                {
                    if(is_string($result)) $result = trim($result);
                }
                if(count(array_filter($answer->result["col$requiredCol"])) <= 0) return array('result' => 'fail', 'callback' => "$('.run-error-msg').removeClass('hidden');");
            }
        }
        if($wizard->type == 'RDA')
        {
            $this->app->loadLang('thinkwizard');
            $run                              = $this->thinkrun->getByID($runID);
            $wizardConfig                     = json_decode($run->options, true);
            $wizardConfig['configureObjects'] = json_decode($wizardConfig['configureObjects'], true);
            $type                             = $wizardConfig['configureObjects']['type'];
            $showObjectName                   = $type == 'product' ? $this->lang->thinkwizard->objects->product : $this->lang->thinkwizard->objects->typeList['RDA'][$type];
            $enableCompetitor                 = !isset($wizardConfig['configureObjects']['enableCompetitor']) || $wizardConfig['configureObjects']['enableCompetitor'] == '1';

            if(empty($enableCompetitor))
            {
                foreach($answer->result["col3"] as &$result){if(is_string($result)) $result = trim($result);}
            }
        }

        foreach($answer->result as $key => $cols)
        {
            if($this->checkColRepeats($cols)) return array('result' => 'fail',  'message' => array("result[$key]" => $this->lang->thinkrun->error->repeatCol), 'callback' => "$('.run-error-msg').addClass('hidden');");
        }

        if($wizard->type == 'RDA' && $enableCompetitor)
        {
            /* RDA类模型启用竟对的校验多列填空题。RDA type models enable multiple column fill in the blank questions with matching validation. */
            if($wizardConfig['configureObjects']['isAssignedObject'] == 1) return false;

            $children      = $this->getFirstNodeQuestion($wizard->id, $runID);
            $firstQuestion = $children[0];
            $firstQuestion->answer = json_decode($firstQuestion->answer, true);

            $subjectName    = $firstQuestion->answer['result'][0];
            $competeNames   = array_filter($answer->result['col1']);
            if(in_array($subjectName, $competeNames)) return array('result' => 'fail', 'message' => array('result[col1]' => str_replace('%s', $showObjectName, $this->lang->thinkwizard->error->objectRepeat)), 'callback' => "$('.run-error-msg').addClass('hidden');");
        }
        elseif($wizard->type == 'RDA' && empty($enableCompetitor))
        {
            /* RDA类模型不启用竟对的校验多列填空题。RDA type models do not enable match checking multi column fill in the blank questions. */
            $competeNames   = array_filter($answer->result['col1']);
            $markets        = array_filter($answer->result['col2']);
            $competesDesc   = $answer->result['col3'];
            $firstCloNames  = sprintf($this->lang->thinkwizard->objects->enterpriseProducts, $showObjectName);

            if(!empty($options->minRows) && count($competeNames) < $options->minRows) return array('result' => 'fail', 'message' => array('result[col1]' => sprintf($this->lang->thinkrun->error->fillInAtLeast, $options->minRows)), 'callback' => "$('.run-error-msg').addClass('hidden');");
            if(!empty($options->minRows) && count($markets) < $options->minRows) return array('result' => 'fail', 'message' => array('result[col2]' => sprintf($this->lang->thinkrun->error->fillInAtLeast, $options->minRows)), 'callback' => "$('.run-error-msg').addClass('hidden');");

            foreach ($competesDesc as $index => $competeDesc)
            {
                if(empty($competeNames[$index]) && !empty($markets[$index])) return array('result' => 'fail', 'message' => array("result[col1][$index]" => sprintf($this->lang->thinkrun->error->empty, $firstCloNames)), 'callback' => "$('.run-error-msg').addClass('hidden');");
                if(!empty($competeNames[$index]) && empty($markets[$index]))  return array('result' => 'fail', 'message' => array("result[col2][$index]" => sprintf($this->lang->thinkrun->error->empty, $this->lang->thinkwizard->objects->market)), 'callback' => "$('.run-error-msg').addClass('hidden');");
                if(!empty($competeDesc) && empty($competeNames[$index]) && empty($markets[$index])) return array('result' => 'fail', 'message' => array("result[col1][$index]" => sprintf($this->lang->thinkrun->error->empty, $firstCloNames), "result[col2][$index]" => sprintf($this->lang->thinkrun->error->empty, $this->lang->thinkwizard->objects->market)), 'callback' => "$('.run-error-msg').addClass('hidden');");
            }
        }
        elseif($wizard->model == 'bcg')
        {
            $competeNames = array_filter($answer->result['col1']);
            if(!empty($options->minRows) && count($competeNames) < $options->minRows) return array('result' => 'fail', 'message' => array('result[col1]' => sprintf($this->lang->thinkrun->error->fillInAtLeast, 2)), 'callback' => "$('.run-error-msg').addClass('hidden');");
        }
        return false;
    }

    /**
     * 检查列数据是否重复。
     * Check if the column data is duplicated.
     *
     * @param  array  $cols
     * @access public
     * @return bool
     */
    public function checkColRepeats($cols)
    {
        $data = array_filter($cols);
        return !$data ? false : (count($data) !== count(array_unique($data)));
    }

    /**
     * 获取多列填空的最多列。
     * get the maximum number of columns for multi-column fill.
     *
     * @param  array  $steps
     * @access public
     * @return int
     */
    public function getMaxColumnResult($steps)
    {
        $maxCount = 1;
        foreach($steps as $item)
        {
            $itemOptions     = is_string($item->options) ? json_decode($item->options) : $item->options;
            $itemFieldsCount = !empty($itemOptions->fields) ? array_search(end($itemOptions->fields), $itemOptions->fields) + 1 : 0;
            if(!empty($itemOptions) && $itemOptions->questionType === 'tableInput' && $maxCount < 2) $maxCount = 2;
            if(!empty($itemOptions) && $itemOptions->questionType === 'multicolumn' && $itemFieldsCount > $maxCount) $maxCount = $itemFieldsCount;
        }
        return $maxCount;
    }

    /**
     * 构建多列填空仅关联列的相关数据。
     * Build a multi column fill in the blank that only associates the relevant data of the column.
     *
     * @param  object $stepLink     分析步骤的关联的相关数据
     * @param  object $stepOptions  分析步骤配置的相关数据
     * @param  array  $result       分析的结果
     * @param  string $block        关联的区块ID
     * @access public
     * @return array
     */
    public function buildColumnStep($stepLink, $stepOptions, $result, $block)
    {
        $columnStep = new stdClass;
        $newField   = array();
        $newResult  = array();
        foreach($stepLink->columnBlock as $columnIndex => $item)
        {
            $fieldIndex = '';
            if($item == $block)
            {
                $fieldIndex                     = $stepLink->column[$columnIndex];
                $newField[$fieldIndex - 1]      = $stepOptions->fields[$fieldIndex - 1];
                $newResult['col' . $fieldIndex] = $result['col' . $fieldIndex];
                /* 关联的顺序不一定从前边到后边的列，需要重新排列让列从前到后展示。*/
                /* The order of association may not necessarily be from front to back columns, and it needs to be rearranged to display columns from front to back. */
                ksort($newField);
            }
        }
        /* 将没有答案的列清除。Clear the columns without answers. */
        foreach($newResult as $resultIndex => $resultItem)
        {
            $resultItem    = (array) $resultItem;
            $newFieldIndex = (int) str_replace('col', '', $resultIndex) - 1;
            $newResultItem = array_filter($resultItem, function($value){ return $value !== '';});
            if(empty($newResultItem)) unset($newField[$newFieldIndex]);
        }

        $columnStep->newField  = $newField;
        $columnStep->newResult = $newResult;
        return $columnStep;
    }

    /**
     * 构建关联区块表格的相关数据。
     * Construct relevant data for the associated block table.
     *
     * @param  array  $questionSteps
     * @access public
     * @return array
     */
    public function buildBlockTableSteps($questionSteps, $blockID)
    {
        $blockTable = array();
        foreach($questionSteps as $step)
        {
            $stepLink     = json_decode($step->link);
            $stepOptions  = $step->options;
            $stepAnswer   = $step->answer;
            $result       = !empty($stepAnswer->result) ? (array) $stepAnswer->result : array();

            if(empty($stepLink)) continue;
            if(empty($result) && $stepOptions->questionType != 'radio' && $stepOptions->questionType != 'checkbox') continue;

            $questionBlocks  = $stepLink->questionBlocks;
            $selectedBlock   = $stepLink->selectedBlock;
            $unSelectedBlock = $stepLink->unSelectedBlock;

            if(!empty($questionBlocks) && $stepOptions->questionType != 'multicolumn' && !empty($result))
            {
                $blockTable[$questionBlocks][] = $step;
                continue;
            }
            /* 多列填空问题，整题关联区块没有填写答案，在结果页面不展示。*/
            /* Multiple column fill in the blank question, with no answer filled in the associated block, will not be displayed on the results page. */
            if(!empty($questionBlocks) && $stepOptions->questionType == 'multicolumn')
            {
                $hidEmptyResult  = false;
                foreach($stepOptions->fields as $index => $field)
                {
                    foreach($result['col' . ($index + 1)] as $resultItem)
                    {
                        if(!empty($resultItem)) $hidEmptyResult = true;
                    }
                }
                if(!empty($hidEmptyResult)) $blockTable[$questionBlocks][] = $step;
                continue;
            }
            if($stepOptions->questionType === 'multicolumn')
            {
                $columnBlock = array_unique($stepLink->columnBlock);

                foreach($columnBlock as $block)
                {
                    $serializedStep  = serialize($step);
                    $newStep         = unserialize($serializedStep);
                    if($stepLink->showMethod == '1')
                    {
                        /* 处理多列填空的数据，只返回有答案的列。Process multi column fill in the blank data and only return columns with answers.*/
                        $newColumnStep            = $this->buildColumnStep($stepLink, $stepOptions, $result, $block);
                        $newStep->options->fields = $newColumnStep->newField;
                        $newStep->answer->result  = $newColumnStep->newResult;
                    }
                    if(!empty($newStep->options->fields)) $blockTable[$block][] = $newStep;
                }
            }
            if($stepOptions->questionType === 'radio' || $stepOptions->questionType === 'checkbox')
            {
                if(!empty($selectedBlock) && !empty($result)) $blockTable[$selectedBlock][] = $step;
                if(!empty($unSelectedBlock))
                {
                    $unSelectedItems = array_diff($stepOptions->fields, $result);
                    $other           = isset($stepOptions->enableOther) && $stepOptions->enableOther == 'on' && empty($stepAnswer->other);
                    /* 存在非选中项的时候，将该题返回。When there are non selected items, return the question. */
                    if(!empty($unSelectedItems) || !empty($other)) $blockTable[$unSelectedBlock][] = $step;
                }
            }
        }
        if(empty($blockTable[$blockID + 1])) return array();
        return $blockTable[$blockID + 1];
    }

    /**
     * 为RDA模型的结果按列分组。
     * Group the results by column for RDA model.
     *
     * @param  array  $questions
     * @param  bool   $enableRatio
     * @access public
     * @return array
     */
    public function groupResultsByColumnForRDA($questions, $enableRatio = true)
    {
        $group = array();
        foreach($questions as $question)
        {
            $question->options = json_decode($question->options, true);
            $question->answer  = json_decode($question->answer, true);
            $result = $question->answer['result'];

            /* 问题计算规则：启用权重的话，问题结果=各个评分题的得分*相应问题权重，否则问题结果=各个评分题的得分。*/
            /* Question Calculation Rule: If weights are enabled, question result = score for each scoring question * corresponding question weight, otherwise question result = score for each scoring question. */
            foreach($result as $key => $value) $result[$key] = $enableRatio ? ((int)$value * (float)$question->options['weight'] / 100) : (int)$value;

            /* 将结果按列分组。 */
            /* Group the results by column. */
            foreach ($result as $key => $value) $group[$key][] = $value;
        }
        return $group;
    }

    /**
     * 为RDA模型计算维度结果。
     * Calculate the result of each dimension for RDA model.
     *
     * @param  array  $data
     * @param  float  $stepWeight
     * @param  bool   $enableRatio
     * @param  bool   $useWeight
     * @access public
     * @return array
     */
    public function calculateDimensionResultForRDA($data, $stepWeight, $enableRatio = true, $useWeight = true)
    {
        /* 维度计算规则：如果启用权重，那么各维度结果=各维度下问题结果之和*相应维度的权重。否则各维度结果=各维度下问题结果的均值*相应维度的权重。*/
        /* Dimension Calculation Rule: If weights are enabled, then each dimension result = sum of the results of the questions under each dimension * weight of the corresponding dimension. Otherwise, the result of each dimension = the mean of the results of the questions under each dimension * the weight of the corresponding dimension. */
        $results = array();
        if($enableRatio)
        {
            foreach($data as $value) $results[] = $useWeight ? array_sum($value) * $stepWeight :  array_sum($value);
        }
        else
        {
            foreach($data as $value)
            {
                $mean = array_sum($value) / count($value);
                $results[] = $useWeight ? $mean * $stepWeight : $mean;
            }
        }
        return $results;
    }

    /**
     * 为RDA模型计算各分析对象的和。
     * Calculate the sum of the object of analysis for RDA model.
     *
     * @param  array  $data
     * @access public
     * @return array
     */
    public function calculateObjectSumForRDA($data)
    {
        if(empty($data)) return array();

        $result = array();
        foreach($data as $item) $result[] = array_sum($item);
        return $result;
    }

    /**
     * 为RDA模型计算各维度结果之和。
     * Calculate sum of the results for each dimension for the RDA model.
     *
     * @param  array  $steps
     * @param  int    $wizardID
     * @param  bool   $enableRatio
     * @param  array  $weights
     * @access public
     * @return array
     */
    public function calculateDimensionSumForRDA($steps, $wizardID, $enableRatio, $weights, $useSums = true)
    {
        $steps      = array_values($steps);
        $stepResult = array();
        $groups     = array();
        foreach($steps as $key => $step)
        {
            $questions = $this->loadModel('thinkstep')->getChildrenByID($wizardID, $step->id);
            $questions = array_filter($questions, function($item) {return $item->type == 'question';});
            $group     = $this->groupResultsByColumnForRDA($questions, $enableRatio);
            $weight    = $enableRatio ? ((float)$weights[$key] / 100) : round(1 / count($steps), 4);

            if($useSums) $stepResult[] = $this->calculateDimensionResultForRDA($group, $weight, $enableRatio);
            $groups[] = $this->calculateDimensionResultForRDA($group, $weight, $enableRatio, false);
        }

        $groups = array_map(null, ...$groups);
        if($useSums)
        {
            $stepResult = array_map(null, ...$stepResult);
            $sums       = $this->calculateObjectSumForRDA($stepResult);
        }
        else
        {
            $sums = array();
        }
        return array('groups' => $groups, 'sums' => $sums);
    }

    /**
     * 获取RDA类模型的分析对象数据。
     * Get the analytic objects data of the RDA model.
     *
     * @param  array  $configureObjects
     * @param  int    $firstStepID
     * @param  int    $wizardID
     * @access public
     * @return array
     */
    public function getObjectsOfRDA($configureObjects, $firstStepID, $wizardID)
    {
        $subjectName = '';
        $competeName = array();
        $markets     = array();
        if($configureObjects['isAssignedObject'] == 1)
        {
            $subjectName = $configureObjects['subjectName'];
            $competeName = array_filter($configureObjects['competeName']);
            if(isset($configureObjects['enableCompetitor']) && $configureObjects['enableCompetitor'] == 0) $markets = array_values(array_filter($configureObjects['markets']));
        }
        else
        {
            $children      = $this->loadModel('thinkstep')->getChildrenByID($wizardID, $firstStepID);
            $children      = array_values(array_filter($children, function($item) {return $item->type == 'question';}));
            if(isset($configureObjects['enableCompetitor']) && $configureObjects['enableCompetitor'] == 0)
            {
                $stepAnswer  = json_decode($children[0]->answer, true);
                $competeName = array_values(array_filter((array)$stepAnswer['result']['col1']));
                $markets     = array_values(array_filter((array)$stepAnswer['result']['col2']));
            }
            else
            {
                $firstQuestion = $children[0];
                $lastQuestion  = end($children);
                $firstQuestion->answer = json_decode($firstQuestion->answer, true);
                $lastQuestion->answer  = json_decode($lastQuestion->answer, true);
                $subjectName = $firstQuestion->answer['result'][0];
                $competeName = array_filter((array)$lastQuestion->answer['result']['col1']);
            }
        }

        return array('subjectName' => $subjectName, 'competeName' => $competeName, 'markets' => $markets);
    }

    /**
     * 获取RDA类模型的维度权重。
     * Get the dimension weights of the RDA model.
     *
     * @param  array  $configureDimension
     * @param  int    $secondStepID
     * @param  int    $wizardID
     * @access public
     * @return array
     */
    public function getWeightsOfRDA($configureDimension, $secondStepID, $wizardID)
    {
        if($configureDimension['enableRatio'] == 0) return array();

        $weights = array();
        if($configureDimension['canConfigureRatio'] == 0)
        {
            $weights = $configureDimension['weights'];
        }
        else
        {
            $steps    = $this->loadModel('thinkstep')->getChildrenByID($wizardID, $secondStepID);
            $steps    = array_values(array_filter($steps, function($item) {return $item->type == 'question';}));
            $question = $steps[0];
            $question->answer = json_decode($question->answer, true);
            $weights = $question->answer['result'];
        }

        return $weights;
    }

    /**
     * 为 RDA 类模型计算各分析对象结果。
     * Calculate results for each analyzed object for RDA model.
     *
     * @param  int    $wizardID
     * @param  int    $runID
     * @param  array  $allSteps
     * @access public
     * @return array
     */
    public function calculateResultForRDA($wizardID, $runID, $allSteps)
    {
        $nodes = array_filter($allSteps, function($item){return $item->parent == 0 && $item->type == 'node';});
        $nodes = array_values($nodes);
        $steps = array_slice($nodes, 2); // 维度数据
        $run   = $this->thinkrun->getByID($runID);
        $run->options = json_decode($run->options, true);
        $configureObjects   = json_decode($run->options['configureObjects'], true);
        $configureDimension = $run->options['configureDimension'];
        $enableRatio        = $configureDimension['enableRatio'] == 1;

        /* 分析模型后的的对象信息和维度权重信息。*/
        /* Object information and dimension weight information after analyzing the model. */
        $objects = $this->getObjectsOfRDA($configureObjects, $nodes[0]->id, $wizardID);
        $weights = $this->getWeightsOfRDA($configureDimension, $nodes[1]->id, $wizardID);

        /* 各分析对象结果。*/
        /* Results for each analyzed object. */
        $enableCompetitor = !isset($configureObjects['enableCompetitor']) || $configureObjects['enableCompetitor'] == 1;
        $result           = $this->calculateDimensionSumForRDA($steps, $wizardID, $enableRatio, $weights, $enableCompetitor);
        $legend           = array();
        $seriesData       = array();
        if(!$enableCompetitor)
        {
            $competeName = array_values($objects['competeName']);
            foreach($objects['markets'] as $index => $item) $legend[] = $item . '/' . $competeName[$index];
            foreach($result['groups'] as $index => $item)   $seriesData[] = array('name' => $legend[$index], 'value' => $item);
        }
        else
        {
            /* 计算最具优势竞品，即各个竞品结果中最大的竞品。 */
            /* Calculate the most dominant competitor, i.e., the largest competitor among the results of each competition. */
            $competitorSum = array_slice($result['sums'], 1);
            $maxKey        = array_keys($competitorSum, max($competitorSum));
            $maxName       = $objects['competeName'][$maxKey[0] + 1];
            $legend        = array($objects['subjectName'], $maxName);

            $this->app->loadLang('thinkwizard');
            $objectType = $configureObjects['type'] == 'product' ? $this->lang->thinkwizard->objects->product : $this->lang->thinkwizard->objects->typeList['RDA'][$configureObjects['type']];
            $subjectName = $this->lang->thinkwizard->objects->subject . $objectType . ': ' . $legend[0];
            $competeName = sprintf($this->lang->thinkrun->bestCompetitor, $objectType) . ': ' . $legend[1];
            $legend      = array($subjectName, $competeName);
            $seriesData  = array(
                array('name' => $legend[0], 'value' => $result['groups'][0],              'label' => array('position' => 'top')),
                array('name' => $legend[1], 'value' => $result['groups'][$maxKey[0] + 1], 'label' => array('position' => 'right')),
            );
        }
        return array('legend' => $legend, 'seriesData' => $seriesData, 'steps' => $steps, 'configureObjects' => $configureObjects);
    }

    /**
     * 给表格中添加结果说明和结论。
     * Add result explanations and conclusions to the table.
     *
     * @param  array  $rows             表格内容数据
     * @param  int    $maxColumnResult  结果列的最大列数
     * @param  string $suggestion       结果说明
     * @param  string $conclusion       结论
     * @param  string $firstColName     首列的key
     * @param  array  $colspan          合并行数据
     * @access public
     * @return array
     */
    public function addSuggestion($rows, $maxColumnResult, $suggestion, $conclusion, $firstColName='index', $colspan=array())
    {
        $blankRow   = array($firstColName => '');
        $suggestion = array($firstColName => $suggestion);
        $conclusion = array($firstColName => $conclusion);

        /* 将添加的空行、结果说明和结论合并单元格。Merge the added blank lines, suggestion, and conclusion into cells. */
        $rowsCount               = count($rows);
        $mergeNum                = 3 + $maxColumnResult;
        $colspan[$rowsCount]     = array('cols' => $firstColName, 'num' => $mergeNum);
        $colspan[$rowsCount + 1] = array('cols' => $firstColName, 'num' => $mergeNum);
        $colspan[$rowsCount + 2] = array('cols' => $firstColName, 'num' => $mergeNum);
        $colspan[$rowsCount + 3] = array('cols' => $firstColName, 'num' => $mergeNum);

        $rows[] = (object) $blankRow;
        $rows[] = (object) $suggestion;
        $rows[] = (object) $blankRow;
        $rows[] = (object) $conclusion;

        $result['rows']    = $rows;
        $result['colspan'] = $colspan;
        return $result;
    }

    /**
     * 构建RDA类模型得分的相关数据。
     * Build relevant data for the $appeals model score.
     *
     * @param  object $run
     * @param  array  $stepsTree
     * @access public
     * @return array
     */
    public function buildScoreOfRDA($run, $stepsTree)
    {
        $result = array();
        $run->options = json_decode($run->options, true);
        $configureObjects   = json_decode($run->options['configureObjects'], true);
        $configureDimension = !empty($run->options['configureDimension']) ? $run->options['configureDimension'] : array();
        $weights = $this->getWeightsOfRDA($configureDimension, $stepsTree[1]->id, $run->wizard);
        $objectList = $this->getObjectsOfRDA($configureObjects, $stepsTree[0]->id, $run->wizard);
        $objects = array_merge(array($objectList['subjectName']), $objectList['competeName']);
        $markets = (isset($configureObjects['enableCompetitor']) && $configureObjects['enableCompetitor'] == '0') ? $objectList['markets'] : array();

        $enableRatio    = $configureDimension['enableRatio'] == 1;
        $dimensionSteps = array_slice($stepsTree, 2);
        $objectsResult  = $this->calculateDimensionSumForRDA($dimensionSteps, $run->wizard, $enableRatio, $weights);

        $result['weights']       = $weights;
        $result['markets']       = array_values(array_filter($markets));
        $result['objects']       = array_values(array_filter($objects));
        $result['objectsResult'] = $objectsResult;
        return $result;
    }

    /**
     * 计算MQA类模型的计算结果。
     * Calculate the result of MQA model.
     *
     * @param  int    $wizardID
     * @param  int    $runID
     * @param  array  $allSteps
     * @access public
     * @return array
     */
    public function calculateResultForMQA($wizardID, $runID, $allSteps)
    {
        /* 获取模型的分析配置。*/
        /* Get the analysis configuration of model. */
        $run = $this->thinkrun->getByID($runID);
        $run->options = json_decode($run->options, true);
        $configureDimension = $run->options['configureDimension'];
        $prerelease         = $run->options['prerelease'];

        /* 获取所有一级节点和产品。*/
        /* Get all first level nodes and products. */
        $nodes         = array_values(array_filter($allSteps, function($item){return $item->parent == 0;}));
        $stepsOfObject = $this->loadModel('thinkstep')->getChildrenByID($wizardID, $nodes[0]->id);
        $stepsOfObject = array_values(array_filter($stepsOfObject, function($item) {return $item->type == 'question';}));
        $stepAnswer    = json_decode($stepsOfObject[0]->answer, true);
        $productNames  = array_values(array_filter($stepAnswer['result']['col1']));

        /* 第二个节点为横坐标 第三个节点为纵坐标. */
        /* The second node is on the horizontal axis, and the third node is on the vertical axis. */
        $xSteps = $this->thinkstep->getChildrenByID($wizardID, $nodes[1]->id);
        $xSteps = array_values(array_filter($xSteps, function($item) {return $item->type == 'question';}));
        $ySteps = $this->thinkstep->getChildrenByID($wizardID, $nodes[2]->id);
        $ySteps = array_values(array_filter($ySteps, function($item) {return $item->type == 'question';}));

        /* 获取的基本计算结果。*/
        /* Get basic calculations. */
        $basicXAxis = $configureDimension['xAxisName'] == 0 ? $this->calculateRelativeMarketShare($xSteps) : $this->calculateAbsoluteMarketShare($xSteps);
        $basicYAxis = $this->calculateSalesGrowthRate($xSteps, $ySteps);

        /* 将计算结果转换成画布需要的格式。*/
        /* Converts the results of the calculation into the format needed for the canvas. */
        $bubbleSizes = $this->calculateBubbleSizes($xSteps);
        $series      = array_map(function($x, $y, $z, $s) {return array($x, $y, $z, $s);}, $basicXAxis, $basicYAxis, $productNames, $bubbleSizes);
        return array('products' => $productNames, 'basicXAxis' => $basicXAxis, 'basicYAxis' => $basicYAxis, 'series'=> $series, 'blockNames' => $prerelease['blocks'], 'runOptions' => $run->options);
    }

    /**
     * 计算气泡大小。
     * Calculate the bubble size.
     *
     * @param  array  $xSteps
     * @access public
     * @return array
     */
    public function calculateBubbleSizes($xSteps)
    {
        $productsAnswer = json_decode($xSteps[0]->answer, true);
        $sales          = $productsAnswer['result'];
        $amountUnit     = $productsAnswer['amountUnit'];
        foreach($sales as $key => &$value) $value = $amountUnit[$key] == 0 ? $value : $value * 10000;

        /* 保持原始键，按销售额排序。*/
        /* Keep original key, sort by sales. */
        uasort($sales, function($a, $b) {return $a <=> $b;});

        /* 获取唯一的销售额值，计算气泡大小, 气泡大小固定最小为36px，按照销售额大小依次增加到1.2倍。*/
        /* Get the unique sales value, calculate the bubble size, the bubble size is fixed to a minimum of 36px, according to the size of the sales increased to 1.2 times in turn. */
        $uniqueSales = array_values(array_unique($sales));
        $bubbleSizes = array();
        foreach($uniqueSales as $key => $sale) $bubbleSizes[$sale] = 36 * (1.2 ** $key);

        $result = array();
        foreach($sales as $key => $sale) $result[$key] = $bubbleSizes[$sale];
        ksort($result);

        return $result;
    }

    /**
     * 构建MQA类模型结果的相关数据。
     * Build relevant result for the MQA model type.
     *
     * @param  object $run
     * @param  array  $stepsTree
     * @access public
     * @return array
     */
    public function buildResultOfMQA($wizardID, $run)
    {
        $run->options         = json_decode($run->options, true);
        $configureDimension   = $run->options['configureDimension'];
        $prerelease           = $run->options['prerelease'];
        $stepsQuestions       = $this->loadModel('thinkwizard')->getStepQuestionsForMQA($wizardID, $run->id);
        $firstChildrenResults = $stepsQuestions['firstChildrenResults'];
        $xSteps               = $stepsQuestions['secondStepQuestions'];
        $ySteps               = $stepsQuestions['thirdStepQuestions'];
        $result               = array();
        $region               = array();

        /* 增长率的计算值和市场占有率的计算值。Calculation values of growth rate and market share. */
        $basicXAxis  = $configureDimension['xAxisName'] == 0 ? $this->calculateRelativeMarketShare($xSteps) : $this->calculateAbsoluteMarketShare($xSteps);
        $basicYAxis  = $this->calculateSalesGrowthRate($xSteps, $ySteps);
        $blocks      = $prerelease['blocks'];
        $xSplitType  = $prerelease['splitType'][1];
        $ySplitType  = $prerelease['splitType'][0];
        $xSplitValue = $xSplitType == '1' ? $prerelease['splitValue'][1] : array(round(array_sum($basicXAxis) / count($basicXAxis) / 100, 2));
        $ySplitValue = $ySplitType == '1' ? $prerelease['splitValue'][0] : array(round(array_sum($basicYAxis) / count($basicYAxis) / 100, 2));

        /* 当前产品上期的销售额、当前产品本期的销售额、细分市场本期销售总额、竞品本期销售总额。*/
        /* The sales volume of the current product in the previous period, the sales volume of the current product in the current period, the total sales volume of the market segment in the current period, and the total sales volume of competing products in the current period. */
        $productLastSalesVolume    = json_decode($ySteps[0]->answer);
        $productCurrentSalesVolume = json_decode($xSteps[0]->answer);
        $markerSalesVolume         = json_decode($xSteps[1]->answer);
        $competitorsSalesVolume    = json_decode($xSteps[2]->answer);

        foreach($basicXAxis as $index => $xAxis)
        {
            $yAxis         = $basicYAxis[$index];
            $xAxisLocation = $this->getAxisLocation($xAxis, $xSplitValue, $configureDimension['xAxisOrder']);
            $yAxisLocation = $this->getAxisLocation($yAxis, $ySplitValue, $configureDimension['yAxisOrder']);
            $regionIndex   = (count($xSplitValue) + 1) * (count($ySplitValue) - $yAxisLocation) + $xAxisLocation + 1;
            $region[]      = $blocks[$regionIndex - 1];
        }
        $result['region']                    = $region;
        $result['basicXAxis']                = $basicXAxis;
        $result['basicYAxis']                = $basicYAxis;
        $result['product']                   = $firstChildrenResults;
        $result['productLastSalesVolume']    = $productLastSalesVolume;
        $result['productCurrentSalesVolume'] = $productCurrentSalesVolume;
        $result['markerSalesVolume']         = $markerSalesVolume;
        $result['competitorsSalesVolume']    = $competitorsSalesVolume;

        return $result;
    }

    /**
     * 获取坐标点的所属位置。
     * Get the location of the coordinate point.
     *
     * @param  float  $axis
     * @param  array  $splitList  分割线数组
     * @param  string $order      排序方式：0 升序 | 1 降序
     * @access public
     * @return int
     */
    public function getAxisLocation($axis, $splitList = array(), $order = '0')
    {
        $axisLocation = 0;
        foreach($splitList as $index => $splitItem)
        {
            $splitValue = $splitItem * 100;
            if($order == '0')
            {
                if($axis < $splitValue)
                {
                    $axisLocation = $index;
                    break;
                }
                if(empty($axisLocation) && $index == count($splitList) - 1 && $axis >= $splitValue) $axisLocation = $index + 1;
            }
            else
            {
                if($axis >= $splitValue)
                {
                    $axisLocation = $index;
                    break;
                }
                if(empty($axisLocation) && $index == count($splitList) - 1 && $axis < $splitValue) $axisLocation = $index + 1;
            }
        }
        return $axisLocation;
    }

    /**
     * 计算相对市场占有率的结果。
     * Calculate the result of relative market share.
     *
     * @param  array  $xSteps
     * @access public
     * @return array
     */
    public function calculateRelativeMarketShare($xSteps)
    {
        /* 计算公式：相对市场占有率 = 产品销售额/最大竞争对手的销售额。*/
        /* Calculation formula: Relative market share = product sales / sales of largest competitor. */
        $products         = $xSteps[0];
        $competitor       = $xSteps[2];
        $productsAnswer   = json_decode($products->answer, true);
        $competitorAnswer = json_decode($competitor->answer, true);
        $unitOfProducts   = $productsAnswer['amountUnit'];
        $unitOfCompetitor = $competitorAnswer['amountUnit'];

        $result = array();
        foreach($productsAnswer['result'] as $key => $value)
        {
            /* 单位转换 0万元 1亿 */
            /* Unit Conversion 0 ten thousand dollars 1 hundred million */
            $value            = $unitOfProducts[$key] == 0 ? $value : $value * 10000;
            $competitiveValue = $unitOfCompetitor[$key] == 0 ? (float)$competitorAnswer['result'][$key] : $competitorAnswer['result'][$key] * 10000;
            $result[] = empty($competitorAnswer['result'][$key]) ? 0 : round($value / $competitiveValue, 4) * 100;
        }

        return $result;
    }

    /**
     * 计算绝对市场占有率的结果。
     * Calculate the result of absolute market share.
     *
     * @param  array  $xSteps
     * @access public
     * @return array
     */
    public function calculateAbsoluteMarketShare($xSteps)
    {
        /* 计算公式：绝对市场占有率 = 产品销售额/所在市场的总销售额。*/
        /* Calculation formula: Absolute market share = Product Sales / Total Sales in the market where it is located. */
        $products       = $xSteps[0];
        $markets        = $xSteps[1];
        $productsAnswer = json_decode($products->answer, true);
        $marketsAnswer  = json_decode($markets->answer, true);
        $unitOfProducts = $productsAnswer['amountUnit'];
        $unitOfMarkets  = $marketsAnswer['amountUnit'];

        $result = array();
        foreach($productsAnswer['result'] as $key => $value)
        {
            $value       = $unitOfProducts[$key] == 0 ? $value : $value * 10000;
            $marketValue = $unitOfMarkets[$key] == 0 ? (float)$marketsAnswer['result'][$key] : $marketsAnswer['result'][$key] * 10000;
            $result[] = empty($marketsAnswer['result'][$key]) ? 0 : round($value / $marketValue, 4) * 100;
        }

        return $result;
    }

    /**
     * 计算销售增长率的结果。
     * Calculate the result of sales growth rate.
     *
     * @param  array  $xSteps
     * @param  array  $ySteps
     * @access public
     * @return array
     */
    public function calculateSalesGrowthRate($xSteps, $ySteps)
    {
        /* 计算公式: 同比销售增长率 = (当前期销售额−去年同期销售额)/去年同期销售额; 环比销售增长率 = (当前期销售额−上一期销售额​)/上一期销售额。*/
        /* Calculation formula: YoY sales growth rate = (current period sales - last year's sales) / last year's sales; QoQ sales growth rate = (current period sales - last period sales) / last period sales. */
        $productsOfLastYear       = $ySteps[0];
        $products                 = $xSteps[0];
        $productsAnswer           = json_decode($products->answer, true);
        $productsOfLastYearAnswer = json_decode($productsOfLastYear->answer, true);
        $unitOfProducts           = $productsAnswer['amountUnit'];
        $unitOfLastYear           = $productsOfLastYearAnswer['amountUnit'];

        $result = array();
        foreach($productsAnswer['result'] as $key => $value)
        {
            $value         = $unitOfProducts[$key] == 0 ? $value : $value * 10000;
            $lastYearValue = $unitOfLastYear[$key] == 0 ? (float)$productsOfLastYearAnswer['result'][$key] : $productsOfLastYearAnswer['result'][$key] * 10000;
            $result[] = empty($lastYearValue) ? 0 : round(($value - $lastYearValue) / $lastYearValue, 4) * 100;
        }

        return $result;
    }

    /**
     * 获取第一个节点下的所有问题。
     * Get all the questions under the first node.
     *
     * @param  int    $wizardID
     * @param  int    $runID
     * @access public
     * @return array
     */
    public function getFirstNodeQuestion($wizardID, $runID)
    {
        $allSteps = $this->loadModel('thinkstep')->getList($wizardID, $runID);
        $nodes    = array_filter($allSteps, function($item){return $item->parent == 0 && $item->type == 'node';});
        $nodes    = array_values($nodes);
        $children = $this->thinkstep->getChildrenByID($wizardID, $nodes[0]->id);
        $children = array_values(array_filter($children, function($item) {return $item->type == 'question';}));
        return $children;
    }

    /**
     * 处理MQA类模型按对象分析表格的相关数据。
     * Build analyze the relevant data of the table by object for the MQA model type.
     *
     * @param  array  $runOptions
     * @param  array  $resultData
     * @access public
     * @return array
     */
    public function processTableDataForMQA($runOptions, $resultData)
    {
        global $lang;

        $dimension      = $runOptions['configureDimension'];
        $parameter      = $runOptions['configureParameter'];
        $newParameter   = $this->loadModel('thinkwizard')->getParameterDateForMQA((object) $parameter);
        $currentDate    = $newParameter->currentDate;
        $lastDate       = $newParameter->lastDate;

        $objectHead['object']              = $lang->thinkrun->tableResult->analysisObject;
        $objectHead['region']              = $lang->thinkrun->tableResult->region;
        $objectHead['lastValue']           = sprintf($lang->thinkrun->tableResult->salesVolume, $lastDate);
        $objectHead['currentValue']        = sprintf($lang->thinkrun->tableResult->salesVolume, $currentDate);
        $objectHead['market']              = $lang->thinkrun->tableResult->belongToMarket;
        $objectHead['marketValue']         = sprintf($lang->thinkrun->tableResult->marketSalesRevenue, $currentDate);
        $objectHead['competitiveProducts'] = $lang->thinkrun->tableResult->competitiveProducts;
        $objectHead['competitorsValue']    = sprintf($lang->thinkrun->tableResult->competitorsSalesRevenue, $currentDate);
        $objectHead['yAxisName']           = $lang->thinkwizard->dimension->yAxisNameList[$dimension['yAxisName']] . '(%)';
        $objectHead['xAxisName']           = $lang->thinkwizard->dimension->xAxisNameList[$dimension['xAxisName']] . '(%)';

        foreach($resultData['region'] as $index => $region)
        {
            $objectBody[] = array(
                'object'              => array_values(array_filter($resultData['product'][0]['col1']))[$index],
                'region'              => $region,
                'lastValue'           => $resultData['productLastSalesVolume']->result[$index] . ' ' . $lang->thinkrun->amountUnit[$resultData['productLastSalesVolume']->amountUnit[$index]],
                'currentValue'        => $resultData['productCurrentSalesVolume']->result[$index] . ' ' . $lang->thinkrun->amountUnit[$resultData['productCurrentSalesVolume']->amountUnit[$index]],
                'market'              => $resultData['product'][1][$index],
                'marketValue'         => $resultData['markerSalesVolume']->result[$index] . ' ' . $lang->thinkrun->amountUnit[$resultData['markerSalesVolume']->amountUnit[$index]],
                'competitiveProducts' => $resultData['product'][2][$index],
                'competitorsValue'    => $resultData['competitorsSalesVolume']->result[$index] . ' ' . $lang->thinkrun->amountUnit[$resultData['competitorsSalesVolume']->amountUnit[$index]],
                'yAxisName'           => (string) round($resultData['basicYAxis'][$index], 2),
                'xAxisName'           => (string) round($resultData['basicXAxis'][$index], 2),
            );
        }
        $object['objectHead'] = $objectHead;
        $object['objectBody'] = $objectBody;
        return $object;
    }

    /**
     * 处理MQA类模型按对象分析表格的相关数据。
     * Build analyze the relevant data of the table by object for the MQA model type.
     *
     * @param  array $runOptions
     * @param  array  $resultData
     * @access public
     * @return array
     */
    public function processTableDataForRDA($runOptions, $resultData, $stepsTree)
    {
        global $lang;
        $configureObjects = json_decode($runOptions['configureObjects'], true);
        $enableCompetitor = !isset($configureObjects['enableCompetitor']) || $configureObjects['enableCompetitor'] == '1';
        $showObjectName   = $configureObjects['type'] == 'product' ? $lang->thinkwizard->objects->product : $lang->thinkwizard->objects->typeList['RDA'][$configureObjects->type];
        $dimensionItem    = array();
        $stepTreeIndex    = 0;
        $objectHead       = array();
        $objectBody       = array();
        $result           = array();

        foreach($stepsTree as $stepTreeIndex => $stepTreeItem)
        {
            if($stepTreeIndex > 1 ) $dimensionItem['dimensionName' . ($stepTreeIndex - 2)] = $stepTreeItem->title . '(' . $resultData['weights'][$stepTreeIndex - 2] . '%)';
            $stepTreeIndex ++;
        }
        $objectHead['object'] = $enableCompetitor ? $lang->thinkrun->tableResult->analysisObject : sprintf($lang->thinkwizard->objects->enterpriseProducts, $showObjectName);
        if(empty($enableCompetitor)) $objectHead['market'] = $lang->thinkrun->market;
        $objectHead = array_merge($objectHead, $dimensionItem);
        $objectHead['sum'] = $lang->thinkrun->tableResult->weightedScore;

        foreach ($resultData['objects'] as $objectIndex => $object)
        {
            $objectBody[$objectIndex]['object'] = $object;
            if(empty($enableCompetitor)) $objectBody[$objectIndex]['market'] = $resultData['markets'][$objectIndex];
            foreach ($resultData['objectsResult']['groups'][$objectIndex] as $groupIndex => $group)
            {
                $objectBody[$objectIndex]['dimensionName' . $groupIndex] = number_format($group, 2);
            }
            $objectBody[$objectIndex]['sum'] = number_format($resultData['objectsResult']['sums'][$objectIndex], 2);
        }

        $result['objectHead'] = $objectHead;
        $result['objectBody'] = $objectBody;
        return $result;
    }
}
