/**
 * 更改导航步骤。
 * Change navigation steps.
 *
 * @access public
 * @returns void
 */
window.changeStepNavs = function(event)
{
    const $a = $(event.target).closest('a');
    if($a.data('disabled')) return;

    const navKey  = $a.data('navKey');
    let modelType = $('.think-config-nav').data('modelType');
    const {currentNav, marketID, wizardID} = $('.think-config-nav').data();

    if(navKey == currentNav) return;

    const submitMethod = currentNav.toLowerCase();
    const jumpMethod   = navKey.toLowerCase();
    modelType          = modelType.toLowerCase();
    const loadUrl      = $.createLink('thinkwizard', `${jumpMethod}for${modelType}`, `marketID=${marketID}&wizardID=${wizardID}`);
    if(submitMethod == 'configureobjects') return submitObjects(loadUrl);
    if(modelType == 'rda' && submitMethod == 'configurequestions') return submitQuestionsOfRDA(loadUrl);
    if(submitMethod == 'configureparameter') return submitParameter(loadUrl);

    const formData = new FormData($('.panel form')[0]);
    $.ajaxSubmit({
        url: $.createLink('thinkwizard', `${submitMethod}for${modelType}`, `marketID=${marketID}&wizardID=${wizardID}`),
        data: formData,
        load: false,
        onSuccess: () => {
            loadPage(loadUrl);
        },
        onFail: (error) => {
            if(error?.message) showValidateMessage(error.message);
        }
    });
}

/**
 * 提交配置对象的数据。
 * Submit data for configuration objects.
 *
 * @access public
 * @returns void
 */
window.submitObjects = function(loadUrl = '')
{
    const formData = new FormData($('.think-objects form')[0]);
    $.ajaxSubmit({
        url: $('.think-objects form').attr('action'),
        data: formData,
        load: false,
        onSuccess: (data) => {
            loadPage(loadUrl ? loadUrl : data.load);
        },
        onFail: (error) => {
            if($('.think-objects .form-batch-table').length)
            {
                $('.form-batch-table .form-control').removeClass('has-error');
                $('.form-batch-table .ajax-form-tip').remove();
            }
            if(error?.message) showValidateMessage(error.message);
        }
    });
}

/**
 * 提交配置问题的数据。
 * Submit data for configuration questions.
 *
 * @access public
 * @returns void
 */
window.submitQuestionsOfRDA = function(loadUrl='')
{
    const formData = new FormData($('form.think-questions-form')[0]);
    $.ajaxSubmit({
        url: $('.form.think-questions-form').attr('action'),
        data: formData,
        onSuccess: () => {
            loadPage(loadUrl);
        },
        onFail: (error) => {
            if($('.think-objects .form-batch-table').length)
            {
                $('.form-batch-table .form-control').removeClass('has-error');
                $('.form-batch-table .ajax-form-tip').remove();
            }
            if(error?.message) showValidateMessage(error.message);
        }
    });
}

/**
 * 提交配置参数的数据。
 * Submit data for configuration objects.
 *
 * @access public
 * @returns void
 */
window.submitParameter = function(loadUrl = '')
{
    const formData = new FormData($('.think-parameter form')[0]);
    $.ajaxSubmit({
        url: $('.think-parameter form').attr('action'),
        data: formData,
        load: false,
        onSuccess: (data) => {
            loadPage(loadUrl ? loadUrl : data.load);
        },
        onFail: (error) => {
            if(error?.message) showValidateMessage(error.message);
        }
    });
}

/**
 * 初始配置对象页面。
 * Initial configuration object page.
 *
 * @access public
 * @returns void
 */
window.initObjects = function()
{
    const isAssignedObject = $('.think-objects').data('isAssignedObject');
    $(".subject-content,.form-batch-table").toggleClass("hidden", isAssignedObject == 0);
}

/**
 * 发布配置模型。
 * Publish configuration model.
 *
 * @access public
 * @returns void
 */
window.publishForPrerelease = function()
{
    const formData = new FormData($('.think-prerelease form')[0]);
    const {marketID, wizardID} = $('.think-config-nav').data();
    $.ajaxSubmit({
        url: $('.think-prerelease form').attr('action'),
        data: formData,
        load: false,
        onSuccess: () => {
            $.ajaxSubmit({url: $.createLink('thinkwizard', 'publish', `marketID=${marketID}&wizardID=${wizardID}&isJump=0`)});
        },
        onFail: (error) => {
            if(error?.message) showValidateMessage(error.message);
        }
    });
}
